#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 28 18:55:02 2023

@author: miguelinacio
"""
#
# Module import
from osgeo import gdal
import numpy as np
import datetime
import os
# Set path
dir_path = os.path.dirname(os.path.realpath(__file__))
#
# Introduction
print('----------------------------------------')
print('            Simplified Marsh            ')
print('             Response Model             ')
print('                   --                   ')
print('                  SMRM                  ')
print('                                        ')
print(' Version: 1.0                           ')
print('----------------------------------------')
print('---------- PARAMETERS TO RUN -----------')
print('----------------------------------------')
print(' 1. Digital Terrain Model (geotiff):    ')
print('    a. Area                             ')
print('    b. Marsh + Tidal Flat               ')
print('    c. RMSE (Root Mean Square Error)    ')
print('       (optional)                       ')
print('                                        ')
print(' 2. Time Range (date):                  ')
print('    a. Initial year                     ')
print('    b. Final year                       ')
print('                                        ')
print(' 3. Transition Heights (m):             ')
print('    a. Tidal Flat / Low Marsh [TF/LM]   ')
print('    b. Low Marsh / High Marsh [LM/HM]   ')
print('    c. High Marsh Limit [+HM]           ')
print('                                        ')
print(' 4. Accretion Rates (mm/year):          ')
print('    a. Tidal Flat (TF)                  ')
print('    b. Low Marsh (LM)                   ')
print('    c. High Marsh (HM)                  ')
print('    d. Acceleration (optional)          ')
print('                                        ')
print(' 5. Sea-level Rise:                     ')
print('    a. Initial Rate (mm/year)           ')
print('    b. Acceleration (mm/year^2)         ')
print('       (optional)                       ')
print('----------------------------------------')
print('                                        ', flush=True)
input('Press any key to continue.              ')
print('                                        ')
#
# Input of parameters
#
# Digital Terrain Model
print('----------------------------------------')
print('-------- DIGITAL TERRAIN MODEL ---------')
print('----------------------------------------', flush=True)
# Open raster files
while True:
    filename_area = input('Area file name: ') # Input area filename
    area_path = os.path.join(dir_path, filename_area + '.tif')
    if os.path.exists(area_path): # If file is found
        print(filename_area + '.tif', 'found.')
        break
    else: # If file is not found
        print(filename_area + '.tif', 'not found.', flush=True)
yearset_area = gdal.Open(area_path) # Open area raster
width_area = yearset_area.RasterXSize # Obtain area raster width
height_area = yearset_area.RasterYSize # Obtain raster area height
# Get the raster data of the area
band_area = yearset_area.GetRasterBand(1)
data_area = band_area.ReadAsArray(0, 0, width_area, height_area)
print('----------------------------------------', flush=True)
while True:
    filename_marsh = input('Salt marsh file name: ') # Input marsh filename
    if not filename_marsh: # If marsh raster is absent
        # SMRM considers MHWS as the marsh terrestrial boundary
        filename_marsh = filename_area
        print('Marsh DTM not found.')
        print('SMRM will consider tidal limit.')
    marsh_path = os.path.join(dir_path, filename_marsh + '.tif')
    if os.path.exists(marsh_path): # If file is found
        print(filename_marsh + '.tif', 'found.')
        break
    else: # If file is not found
        print(filename_marsh + '.tif', 'not found. Please try again.', \
              flush=True)
yearset_marsh = gdal.Open(marsh_path) # Open marsh raster
width_marsh = yearset_marsh.RasterXSize # Obtain marsh raster width
height_marsh = yearset_marsh.RasterYSize # Obtain marsh raster height
# Get the raster data of the marsh
band_marsh = yearset_marsh.GetRasterBand(1)
data_marsh = band_marsh.ReadAsArray(0, 0, width_marsh, height_marsh)
print('----------------------------------------', flush=True)
# Root Mean Square Error (RMSE)
RMSE = input('RMSE (m): ') # Input the RMSE
if not RMSE: # if RMSE is absent, no error is considered
    print('RMSE = 0.0 m')
    RMSE = 0.0
else:
    RMSE = float(RMSE)
print('----------------------------------------')
print('                                        ')

# Time range
print('----------------------------------------')
print('-------------- TIME RANGE --------------')
print('----------------------------------------', flush=True)
present_year = datetime.datetime.now().year # Obtain the present year
while True:
    initial_year = input('Initial year: ') # Input initial year
    if not initial_year: # If final year is absent
        print('Initial year =', present_year)
        initial_year = present_year
        break
    elif float(initial_year) > present_year: # If initial year > present
        print("The year must be", present_year, "or earlier.", flush=True)
    else:
        initial_year = int(initial_year)
        break 
print('----------------------------------------', flush=True)
while True:
    final_year = input('Final year: ') # Input final year
    if not final_year:
        print('Input is required. Please try again.', flush=True)
    elif float(final_year) <= float(initial_year) or float(final_year) > 2100:
        # If final year > initial year
        print("Final year must be between", initial_year+1 , "and 2100.", \
              flush=True)
    else:
        final_year = int(final_year)
        break
print('----------------------------------------')
print('                                        ')
# Transition heights (local tidal levels)
print('----------------------------------------')
print('---------- TRANSITION HEIGHTS ----------')
print('----------------------------------------', flush=True)       
while True:
    MHWNi = input('Transition level [TF/LM] (m): ') # Input TF/LM (MHWN)
    if MHWNi:
        MHWNi = float(MHWNi)
        break
    else:
        print("Input is required. Please try again.", flush=True)
print('----------------------------------------', flush=True)
while True:
    MHWi = input('Transition level [LM/HM] (m): ') # Input LM/HM (MHW)
    if MHWi:
        MHWi = float(MHWi)
        break
    else:
        print("Input is required. Pleasy try again.", flush=True)
print('----------------------------------------', flush=True)
while True:
    MHWSi = input('High marsh limit [+HM] (m): ') # Input +HM (MHWS)
    if MHWSi:
        MHWSi = float(MHWSi)
        break
    else:
        print("Input is required. Pleasy try again.", flush=True)

print('----------------------------------------')
print('                                        ')
# Accretion rates
print('----------------------------------------')
print('----------- ACCRETION RATES ------------')
print('----------------------------------------', flush=True)
Rate_TFi = input('Tidal Flat [TF] (mm/year): ') # Input TF accretion rate
if not Rate_TFi:
    print('[TF] = 0.0 mm/year')
    Rate_TFi = 0.0
else:
    Rate_TFi = float(Rate_TFi) / 1000
Acc_TF = input('[TF] Acceleration (mm/year^2): ') # Input TF acceleration
if not Acc_TF: # TF acceleration is an optional parameter
    print('Acceleration = 0.0 mm/year^2')
    Acc_TF = 0.0
else:
    Acc_TF = float(Acc_TF) / 1000
print('----------------------------------------', flush=True)
Rate_LMi = input('Low Marsh [LM] (mm/year): ') # Input LM accretion rate
if not Rate_LMi:
    print('[LM] = 0.0 mm/year')
    Rate_LMi = 0.0
else:
    Rate_LMi = float(Rate_LMi) / 1000
Acc_LM = input('[LM] Acceleration (mm/year^2): ') # Input LM acceleration
if not Acc_LM: # LM acceleration is an optional parameter
    print('Acceleration = 0.0 mm/year^2')
    Acc_LM = 0.0
else:
    Acc_LM = float(Acc_LM) / 1000
print('----------------------------------------', flush=True)
Rate_HMi = input('High Marsh [HM] (mm/year): ') # Input HM accretion rate
if not Rate_HMi:
    print('[HM] = 0.0 mm/year')
    Rate_HMi = 0.0
else:
    Rate_HMi = float(Rate_HMi) / 1000
Acc_HM = input('[HM] Acceleration (mm/year^2): ') # Input HM acceleration
if not Acc_HM: # HM acceleration is an optional parameter
    print('Acceleration = 0.0 mm/year^2')
    Acc_HM = 0.0
else:
    Acc_HM = float(Acc_HM) / 1000
print('----------------------------------------')
print('                                        ')
# Sea-level rise (SLR)
print('----------------------------------------')
print('------------ SEA-LEVEL RISE ------------')
print('----------------------------------------', flush=True)
while True:
    SLR_rate = input('Initial rate (mm/year): ') # Input initial SLR rate
    if SLR_rate: # Convert mm/year to m/year
        SLR_rate = float(SLR_rate) / 1000
        break
    else:
        print("Input is required. Please try again.", flush=True)
print('----------------------------------------', flush=True)
SLR_acc = input('Acceleration (mm/year^2): ') # Input constant acceleration
if not SLR_acc: # SLR acceleration is an optional parameter
    print('Acceleration = 0.0 mm/year^2')
    SLR_acc = 0.0
else: # Convert mm/year^2 to m/year^2
    SLR_acc = float(SLR_acc) / 1000
print('----------------------------------------')
#
# Define vectors
SLR_initial = SLR_rate * 1000 # Initial SLR rate (mm/year)
years = list(range(initial_year+1,final_year+1)) # Years vector
MHWNi = MHWNi + RMSE # Add RMSE to MHWN for the initial year
MHWi = MHWi + RMSE # Add RMSE to MHW for the initial year
MHWSi = MHWSi + RMSE # Add RMSE to MHWS for the initial year
MHWN = [MHWNi] # Define MHWN vector
MHW = [MHWi] # Define MHW vector
MHWS = [MHWSi] # Define MHWS vector
Rate_TF = [Rate_TFi] # Define TF accretion rate vector
Rate_LM = [Rate_LMi] # Define LM accretion rate vector
Rate_HM = [Rate_HMi] # Define HM accretion rate vector
# Create the previous vectors for the define time interval
for i in range(0,len(years)):
    MHWNi = MHWNi + SLR_rate # Calculate MHWN for each year
    MHWi = MHWi + SLR_rate # Calculate MHW for each year
    MHWSi = MHWSi + SLR_rate # Calculate MHWS for each year
    Rate_TFi = Rate_TFi + Acc_TF # Calculate TF accretion rate for each year
    Rate_LMi = Rate_LMi + Acc_LM # Calculate LM accretion rate for each year
    Rate_HMi = Rate_HMi + Acc_HM # Calculate HM accretion rate for each year
    SLR_rate = SLR_rate + SLR_acc # Calculate SLR rate for each year
    MHWN.append(MHWNi) # Append the calculated MHWN to the MHWN vector
    MHW.append(MHWi) # Append the calculated MHW to the MWH vector
    MHWS.append(MHWSi) # Append the calculated NHWS to the MHWS vector
    Rate_TF.append(Rate_TFi) # Append the calculated TF rate to the TF vector
    Rate_LM.append(Rate_LMi) # Append the calculated LM rate to the LM vector
    Rate_HM.append(Rate_HMi) # Append the calculated HM rate to the HM vector
SLR_final = SLR_rate * 1000 # Final SLR rate (mm/year)
print('                                        ')
# Input parameters confirmation
print('----------------------------------------')
print('---- INPUT PARAMETERS CONFIRMATION -----')
print('----------------------------------------')
print(' - Initial marsh:', filename_marsh + '.tif') # Marsh filename
print(' - Area:', filename_area + '.tif') # Area filename
print('                                        ')
print(' - Years:', initial_year, '-', final_year) # Time interval
# SLR rates interval (initial - final)
print(' - SLR Rate:', "{:.3f}".format(SLR_initial), '-', \
      "{:.3}".format(SLR_final), 'mm/year')
print('                                        ')
print('   Transition') # Transition heights
print(' | Heights | ', initial_year, ' |')
print(' | [TF/LM] |',"{:.2f}".format(MHWN[0] - RMSE), 'm |') # TF/LM
print(' | [LM/HM] |',"{:.2f}".format(MHW[0] - RMSE),'m |') # LM/HM
print(' |  [+HM]  |',"{:.2f}".format(MHWS[0] - RMSE), 'm |') # HM limit
print('                                        ')
print('   Accretion    Rates') # Accretion rates
print(' | Env. | mm/year | m/year |')
print(' |  TF  | ',"{:.2f}".format(Rate_TF[0]*1000),\
      '  |',"{:.4f}".format(Rate_TF[0]),'|')
print(' |  LM  | ',"{:.2f}".format(Rate_LM[0]*1000),\
      '  |',"{:.4f}".format(Rate_LM[0]),'|')
print(' |  HM  | ',"{:.2f}".format(Rate_HM[0]*1000),\
      '  |',"{:.4f}".format(Rate_HM[0]),'|')
print('----------------------------------------')
print('                                        ', flush=True)
input('Press Enter to continue.                ')
print('                                        ')
print('----------------------------------------')
print('                                        ')
print("\rProcessing: 0% completed", end='')
# Define background values as NaN (Not a Number)
for i in range(len(data_area)):
    for j in range(len(data_area[i])):
        if data_area[i,j] < -100 or data_area[i,j] > 100:
            data_area[i,j] = np.nan
        if data_marsh[i,j] < -100 or data_marsh[i,j] > 100:
            data_marsh[i,j] = np.nan
#
original_data = data_area.copy() # Creates a copy of the original DTM
#
# Simulation:
for t in range(0, len(years)): # For cycle for each year
    percentage_completed = (t / len(years)) * 100 # Percentage completed
    print("\rProcessing: {:.0f}% completed".format(percentage_completed),\
          end='')
    for i in range(len(data_area)): # Application of accredtion rates
        for j in range(len(data_area[i])):
            # When pixel is above MHWS
            # No accretion rate is applied
            if data_area[i,j] > MHWS[t] or np.isnan(data_area[i,j]):
                data_area[i,j] = data_area[i,j]
            # When pixel is between MHW and MHWS
            # HM accretion rate is applied
            elif data_area[i,j] > MHW[t]:
                data_area[i,j] = data_area[i,j] + Rate_HM[t]
            # When pixel is between MHWN and MHW
            # LM accretion rate is applied
            elif data_area[i,j] > MHWN[t]:
                data_area[i,j] = data_area[i,j] + Rate_LM[t]
            # When pixel is below MHWN
            # TF accretion rate is applied
            elif data_area[i,j] <= MHWN[t]:
                data_area[i,j] = data_area[i,j] + Rate_TF[t]
            else:
                print('Error!')
    # Check if the simulated year is multiple of 10 or the final year
    if years[t] % 10 == 0 or  years[t] == final_year:
        mdt_alt = np.zeros((height_area,width_area))
        # Check what are the pixels that were changed during the process
        for k in range(len(data_area)):
            for l in range(len(data_area[k])):
                if data_area[k,l] == original_data[k,l] or \
                    (np.isnan(data_area[k,l]) and \
                     np.isnan(original_data[k,l])):
                    mdt_alt[k,l] = 0
                else:
                    mdt_alt[k,l] = 1
        # Production of the result
        mdt_new = np.full((height_area,width_area), np.nan)  
        # Use the new generated DTM for the pixels that were changed
        # Use the original DTM for the areas that were not changed        
        for m in range(len(data_area)): 
            for n in range(len(data_area[m])):
                if mdt_alt[m,n] == 0:
                    mdt_new[m,n] = data_marsh[m,n]
                elif mdt_alt[m,n] == 1:
                    mdt_new[m,n] = data_area[m,n]
                else:
                    print('Error!')
        output_name = "SMRM_" + str(years[t]) + ".tif" # Output filename
        # Export the results
        driver = gdal.GetDriverByName('GTiff')
        output_ds = driver.CreateCopy(output_name, yearset_area)
        output_band = output_ds.GetRasterBand(1)
        output_band.WriteArray(mdt_new)
# Close the yearsets
yearset_area = None
yearset_marsh = None
output_ds = None
print("\rProcessing: 100% completed")
print('                                        ')
# Presentation of the results
# Tidal table for each decade
print('----------------------------------------')
print('--------------- RESULTS ----------------')
print('----------------------------------------')
print(' | Year | [TF/LM] | [LM/HM] | [+HM] | m')
print(' |', initial_year, '| ', "{:.2f}".format(MHWN[0] - RMSE),\
      '  | ', "{:.2f}".format(MHW[0] - RMSE),\
          '  |', "{:.2f}".format(MHWS[0] - RMSE),' | m')
for i in range(0, len(years)):
    if years[i] % 10 == 0 or years[i] == final_year:
        print(' |', years[i], '| ', "{:.2f}".format(MHWN[i] - RMSE),\
              '  | ', "{:.2f}".format(MHW[i] - RMSE),\
                  '  |', "{:.2f}".format(MHWS[i] - RMSE),' | m')
print(' - New DTMs were exported.                 ')
print('----------------------------------------')
print('Thank you for using SMRM.               ')
print('                                        ')
print('E-mail: mfinacio@fc.ul.pt               ')